from time import time_ns
from typing import Collection, Dict, List, Optional, Tuple, Union, cast
import logging

from sarus_data_spec.attribute import attach_properties
from sarus_data_spec.constants import VARIANT_UUID
from sarus_data_spec.context import global_context
from sarus_data_spec.dataset import transformed
from sarus_data_spec.dataspec_validator.typing import DataspecValidator
from sarus_data_spec.manager.ops.processor import routing
from sarus_data_spec.scalar import privacy_budget
from sarus_data_spec.variant_constraint import (
    dp_constraint,
    mock_constraint,
    syn_constraint,
)
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st

logger = logging.getLogger(__name__)


def attach_variant(
    original: st.DataSpec,
    variant: st.DataSpec,
    kind: st.ConstraintKind,
) -> None:
    attach_properties(
        original,
        properties={
            # TODO deprecated in SDS >= 2.0.0 -> use only VARIANT_UUID
            kind.name: variant.uuid(),
            VARIANT_UUID: variant.uuid(),
        },
        name=kind.name,
    )


def rewrite(
    dataspec_validator: DataspecValidator,
    dataspec: st.DataSpec,
    kind: st.ConstraintKind,
    public_context: Collection[str],
    privacy_limit: Optional[st.PrivacyLimit],
) -> Optional[st.DataSpec]:
    """Returns a compliant Node or None."""

    if kind == st.ConstraintKind.SYNTHETIC:
        variant, _ = rewrite_synthetic(
            dataspec_validator,
            dataspec,
            public_context,
        )
        return variant

    elif kind == st.ConstraintKind.MOCK:
        mock_variant, _ = rewrite_mock(
            dataspec_validator,
            dataspec,
            public_context,
        )
        return mock_variant

    if privacy_limit is None:
        raise ValueError(
            "Privacy limit must be defined for PUP or DP rewriting"
        )

    if kind == st.ConstraintKind.DP:
        variant, _ = rewrite_dp(
            dataspec_validator,
            dataspec,
            public_context=public_context,
            privacy_limit=privacy_limit,
        )
        return variant

    elif kind == st.ConstraintKind.PUP:
        raise NotImplementedError("PUP rewriting")

    else:
        raise ValueError(
            f"Privacy policy {kind} rewriting not implemented yet"
        )


def rewrite_synthetic(
    dataspec_validator: DataspecValidator,
    dataspec: st.DataSpec,
    public_context: Collection[str],
) -> Tuple[st.DataSpec, Collection[str]]:
    # Current dataspec verifies the constraint?
    for constraint in dataspec_validator.verified_constraints(dataspec):
        if dataspec_validator.verifies(
            constraint,
            st.ConstraintKind.SYNTHETIC,
            public_context,
            privacy_limit=None,
        ):
            return dataspec, public_context

    # Current dataspec has a variant that verifies the constraint?
    for variant in dataspec.variants():
        if variant is None:
            logger.info(f"Found a None variant for dataspec {dataspec.uuid()}")
            continue
        for constraint in dataspec_validator.verified_constraints(variant):
            if dataspec_validator.verifies(
                constraint,
                st.ConstraintKind.SYNTHETIC,
                public_context,
                privacy_limit=None,
            ):
                return variant, public_context

    # Derive the SD from the parents SD
    if dataspec.is_transformed():
        transform = dataspec.transform()
        args, kwargs = dataspec.parents()

        syn_args: List[Union[st.DataSpec, st.Transform]] = [
            rewrite_synthetic(dataspec_validator, arg, public_context)[0]
            if isinstance(arg, st.DataSpec)
            else arg
            for arg in args
        ]
        syn_kwargs: Dict[str, Union[st.DataSpec, st.Transform]] = {
            name: rewrite_synthetic(dataspec_validator, arg, public_context)[0]
            if isinstance(arg, st.DataSpec)
            else arg
            for name, arg in kwargs.items()
        }

        syn_variant = cast(
            st.DataSpec,
            transformed(
                transform,
                *syn_args,
                dataspec_type=sp.type_name(dataspec.prototype()),
                dataspec_name=None,
                **syn_kwargs,
            ),
        )
        syn_constraint(
            dataspec=syn_variant, required_context=list(public_context)
        )
        attach_variant(dataspec, syn_variant, kind=st.ConstraintKind.SYNTHETIC)
        return syn_variant, public_context

    elif dataspec.is_public():
        return dataspec, public_context
    else:
        raise TypeError(
            "Non public source Datasets cannot"
            "be rewritten to Synthetic, a synthetic variant"
            "should have been created downstream in the graph."
        )


def rewrite_mock(
    dataspec_validator: DataspecValidator,
    dataspec: st.DataSpec,
    public_context: Collection[str],
) -> Tuple[Optional[st.DataSpec], Collection[str]]:
    """rewrite the MOCK variant of a DataSpec.

    Note that the MOCK rewriting only makes sense for internally transformed
    dataspecs. For externally transformed dataspecs, the MOCK is computed
    before the dataspec, so we can only fetch it.
    """
    for constraint in dataspec_validator.verified_constraints(dataspec):
        if dataspec_validator.verifies(
            constraint,
            st.ConstraintKind.MOCK,
            public_context,
            privacy_limit=None,
        ):
            return dataspec, public_context

    # Current dataspec has a variant that verifies the constraint?
    for variant in dataspec.variants():
        if variant is None:
            logger.info(f"Found a None variant for dataspec {dataspec.uuid()}")
            continue
        for constraint in dataspec_validator.verified_constraints(variant):
            if dataspec_validator.verifies(
                constraint,
                st.ConstraintKind.MOCK,
                public_context,
                privacy_limit=None,
            ):
                return variant, public_context

    if not dataspec.is_transformed():
        if not dataspec.is_public():
            raise ValueError(
                "Cannot rewrite the MOCK of a non public source DataSpec. "
                "A MOCK should be set manually downstream in the "
                "computation graph."
            )
        else:
            return dataspec, public_context

    # The DataSpec is the result of an internal transform
    transform = dataspec.transform()
    args, kwargs = dataspec.parents()
    mock_args = [
        arg.variant(st.ConstraintKind.MOCK)
        if isinstance(arg, st.DataSpec)
        else arg
        for arg in args
    ]
    named_mock_args = {
        name: arg.variant(st.ConstraintKind.MOCK)
        if isinstance(arg, st.DataSpec)
        else arg
        for name, arg in kwargs.items()
    }
    if any([m is None for m in mock_args]) or any(
        [m is None for m in named_mock_args.values()]
    ):
        raise ValueError(
            f"Cannot derive a mock for {dataspec} "
            "because of of the parent has a None MOCK."
        )

    typed_mock_args = [cast(st.DataSpec, ds) for ds in mock_args]
    typed_named_mock_args = {
        name: cast(st.DataSpec, ds) for name, ds in named_mock_args.items()
    }

    mock: st.DataSpec = transformed(
        transform,
        *typed_mock_args,
        dataspec_type=sp.type_name(dataspec.prototype()),
        dataspec_name=None,
        **typed_named_mock_args,
    )
    mock_constraint(mock)
    attach_variant(dataspec, mock, st.ConstraintKind.MOCK)

    return mock, public_context


def rewrite_dp(
    dataspec_validator: DataspecValidator,
    dataspec: st.DataSpec,
    public_context: Collection[str],
    privacy_limit: st.PrivacyLimit,
) -> Tuple[st.DataSpec, Collection[str]]:
    """Simple DP rewriting.

    Only check the dataspec's parents, do not go further up in the graph.
    """
    # Current dataspec verifies the constraint?
    for constraint in dataspec_validator.verified_constraints(dataspec):
        if dataspec_validator.verifies(
            variant_constraint=constraint,
            kind=st.ConstraintKind.DP,
            public_context=public_context,
            privacy_limit=privacy_limit,
        ):
            return dataspec, public_context

    # Current dataspec has a variant that verifies the constraint?
    for variant in dataspec.variants():
        for constraint in dataspec_validator.verified_constraints(variant):
            if dataspec_validator.verifies(
                variant_constraint=constraint,
                kind=st.ConstraintKind.DP,
                public_context=public_context,
                privacy_limit=privacy_limit,
            ):
                return variant, public_context

    if not dataspec.is_transformed():
        return rewrite_synthetic(dataspec_validator, dataspec, public_context)

    # Check that there is a positive epsilon
    delta_epsilon_dict = privacy_limit.delta_epsilon_dict()
    if len(delta_epsilon_dict) == 1:
        epsilon = list(delta_epsilon_dict.values()).pop()
        if epsilon == 0:
            return rewrite_synthetic(
                dataspec_validator, dataspec, public_context
            )

    transform = dataspec.transform()
    if dataspec.prototype() == sp.Dataset:
        dataset = cast(st.Dataset, dataspec)
        _, DatasetStaticChecker = routing.get_dataset_op(transform)
        is_dp_writable = DatasetStaticChecker(dataset).is_dp_writable(
            public_context
        )
        dp_transform = DatasetStaticChecker(dataset).dp_transform()
    else:
        scalar = cast(st.Scalar, dataspec)
        _, ScalarStaticChecker = routing.get_scalar_op(transform)
        is_dp_writable = ScalarStaticChecker(scalar).is_dp_writable(
            public_context
        )
        dp_transform = ScalarStaticChecker(scalar).dp_transform()

    if not is_dp_writable:
        return rewrite_synthetic(dataspec_validator, dataspec, public_context)

    # Create the DP variant
    assert dp_transform is not None
    budget = privacy_budget(privacy_limit)
    seed = global_context().generate_seed(salt=time_ns())

    args, kwargs = dataspec.parents()
    dp_variant = cast(
        st.DataSpec,
        transformed(
            dp_transform,
            *args,
            dataspec_type=sp.type_name(dataspec.prototype()),
            dataspec_name=None,
            budget=budget,
            seed=seed,
            **kwargs,
        ),
    )
    dp_constraint(
        dataspec=dp_variant,
        required_context=list(public_context),
        privacy_limit=privacy_limit,
    )
    attach_variant(
        original=dataspec,
        variant=dp_variant,
        kind=st.ConstraintKind.DP,
    )

    # We also attach the dataspec's synthetic variant to be the DP dataspec's
    # synthetic variant. This is to avoid to have DP computations in the MOCK.
    syn_variant = dataspec.variant(st.ConstraintKind.SYNTHETIC)
    if syn_variant is None:
        raise ValueError("Could not find a synthetic variant.")
    attach_variant(
        original=dp_variant,
        variant=syn_variant,
        kind=st.ConstraintKind.SYNTHETIC,
    )

    return dp_variant, public_context



from typing import Any, Optional, Union

import numpy.typing as npt
import sarus_data_spec.typing as st
from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from scipy.sparse import spmatrix
    from sklearn.base import BaseEstimator
except ModuleNotFoundError:
    BaseEstimator = Any
    spmatrix = Any


class ReturnSelfExternalOpImplementation(ExternalOpImplementation):
    """Trait with its implementation of `py_out_hint`.

    It returns the type of the parameter `"this"` in the op signature, which
    should be part of the transform definition.
    """

    def py_output_hint(
        self,
        transform: st.Transform,
        *arguments: Union[st.DataSpec, st.Transform],
        **named_arguments: Union[st.DataSpec, st.Transform],
    ) -> Optional[str]:
        all_args = self.signature().bind(
            transform, *arguments, **named_arguments
        )
        this = all_args["this"]
        return this.python_type()


class sk_fit(ReturnSelfExternalOpImplementation):
    _transform_id = "sklearn.SK_FIT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
        SarusParameter(
            name="y",
            annotation=Optional[Union[npt.ArrayLike, spmatrix]],
            default=None,
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        if kwargs["sample_weight"] is None:
            del kwargs["sample_weight"]
        fitted_model = this.fit(**kwargs)
        return fitted_model


class sk_fit_y(ReturnSelfExternalOpImplementation):
    _transform_id = "sklearn.SK_FIT_Y"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="y",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        fitted_model = this.fit(**kwargs)
        return fitted_model


class sk_predict(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PREDICT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.predict(**kwargs)


class sk_predict_callable(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PREDICT_CALLABLE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.predict


class sk_predict_log_proba(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PREDICT_LOG_PROBA"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.predict_log_proba(**kwargs)


class sk_predict_log_proba_callable(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PREDICT_LOG_PROBA_CALLABLE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.predict_log_proba


class sk_predict_proba(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PREDICT_PROBA"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.predict_proba(**kwargs)


class sk_predict_proba_callable(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PREDICT_PROBA_CALLABLE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.predict_proba


class sk_transform(ExternalOpImplementation):
    _transform_id = "sklearn.SK_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, X) = signature.collect_args()
        return this.transform(X)


class sk_inverse_transform(ExternalOpImplementation):
    _transform_id = "sklearn.SK_INVERSE_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, X = signature.collect_args()
        return this.inverse_transform(X)


class sk_pipeline_fit(ReturnSelfExternalOpImplementation):
    _transform_id = "sklearn.SK_PIPELINE_FIT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
        SarusParameter(
            name="y",
            annotation=Optional[Union[npt.ArrayLike, spmatrix]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        fitted_model = this.fit(**kwargs)
        return fitted_model


class sk_pipeline_fit_transform(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PIPELINE_FIT_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
        SarusParameter(
            name="y",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.fit_transform(**kwargs)


class sk_pipeline_transform(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PIPELINE_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, X) = signature.collect_args()
        return this.transform(X)


class sk_score(ExternalOpImplementation):
    _transform_id = "sklearn.SK_SCORE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
        SarusParameter(
            name="y",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.score(**kwargs)


class sk_fit_transform(ExternalOpImplementation):
    _transform_id = "sklearn.SK_LABEL_ENCODER_FIT_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
        SarusParameter(
            name="y",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.fit_transform(**kwargs)


class sk_split(ExternalOpImplementation):
    _transform_id = "sklearn.SK_SPLIT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
        SarusParameter(
            name="y",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="groups",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.split(**kwargs)


class sk_get_n_splits(ExternalOpImplementation):
    _transform_id = "sklearn.SK_GET_N_SPLITS"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.get_n_splits(**kwargs)



import typing as t

import pyarrow as pa

from sarus_data_spec.transform import (
    has_sarus_default_output,
    extract_sarus_default_output,
)
from sarus_data_spec.manager.async_utils import (
    async_iter,
)
from sarus_data_spec.manager.ops.base import (
    _ensure_batch_correct,
)
import sarus_data_spec.typing as st
from sarus_data_spec.status import DataSpecErrorStatus


async def check_first_batch_empty(
    ait: t.AsyncIterator[pa.RecordBatch],
) -> t.Tuple[bool, t.AsyncIterator[pa.RecordBatch]]:
    """Returns False if the async iterator have a first not empty batch, and return the original iterator. R"""
    try:
        first_item = await ait.__anext__()

        async def chained_iterator() -> t.AsyncIterator[pa.RecordBatch]:
            yield first_item
            async for item in ait:
                yield item

        return first_item.num_rows == 0, chained_iterator()
    except StopAsyncIteration:
        return True, ait


async def ensure_batch_correct_and_not_empty(
    dataset: st.DataSpec,
    async_iterator: t.AsyncIterator[pa.RecordBatch],
    batch_size: int,
) -> t.AsyncIterator[pa.RecordBatch]:
    """Ensures that the provided asynchronous iterator over a dataset does not yield empty batches.

    This function checks if the iterator is initially empty. If not, it simply returns the original iterator.
    If the iterator is empty and has a default output defined (i.e., 'sarus_default_output').
    For datasets with a default output, it creates and returns a new async iterator that yields batches created from the default output.
    If no default output is present, it raises a ValueError indicating that the computation results in an empty dataset without a schema,
    thus limiting the transformations that can be applied."""

    async def identity(x: pa.Array) -> pa.Array:
        return x

    ait = _ensure_batch_correct(
        async_iterator=async_iterator,
        func_to_apply=identity,
        batch_size=batch_size,
    )

    is_empty, original_async_iterator = await check_first_batch_empty(ait)
    if not is_empty:
        return original_async_iterator
    # use default value if it exists
    elif has_sarus_default_output(dataset.transform()):
        if not dataset.is_pup():
            default_table = extract_sarus_default_output(dataset.transform())
            return async_iter(
                default_table.to_batches(max_chunksize=batch_size)
            )
        else:
            raise NotImplementedError(
                "Sarus default output is not implemented for pup dataset"
            )
    else:
        message_empty_ait = f"""Warning: The computation of dataspec: {dataset.uuid()} returns an empty result, which does not have a schema. This limits the transformations you can apply to the resulting object.
    If you believe that the computation on real data is not empty, you may provide the schema by adding a dummy example using the parameter 'sarus_default_output'.
    You can find more details at https://docs.sarus.tech/User%20Guide/data_science_%26_AI.html#handling-empty-mocks-error. """
        raise DataSpecErrorStatus(
            (
                False,
                message_empty_ait,
            )
        )



from collections import defaultdict
import logging
import typing as t
from sarus_data_spec.manager.ops.sql_utils.pyqrlew_utils import (
    SARUS_TO_QRLEW_DIALECT_MAPPING,
)
import sarus_data_spec.type as sdt
import sarus_data_spec.typing as st
import pyqrlew as pyqrl


logger = logging.getLogger(__name__)


def flatten_queries_dict(
    queries: st.NestedQueryDict,
) -> t.Dict[t.Tuple[str, ...], str]:
    """Transform nested dict in linear dict where each
    key is the tuple of the nesting path"""

    final_dict: t.Dict[t.Tuple[str, ...], str] = {}

    def update_dict(
        curr_path: t.List[str],
        dict_to_update: t.Dict[t.Tuple[str, ...], t.Any],
        query_or_dict: t.Union[t.Dict[str, t.Any], t.Any],
    ) -> None:
        if isinstance(query_or_dict, dict):
            for name, sub_query in query_or_dict.items():
                update_dict(
                    curr_path=[*curr_path, name],
                    dict_to_update=dict_to_update,
                    query_or_dict=sub_query,
                )
        else:
            dict_to_update[tuple(curr_path)] = t.cast(str, query_or_dict)
        return

    for name, query_or_dict in queries.items():
        update_dict(
            query_or_dict=query_or_dict,
            curr_path=[name],
            dict_to_update=final_dict,
        )
    return final_dict


def nest_queries(
    queries: t.Dict[t.Tuple[str, ...], str],
) -> st.NestedQueryDict:
    """It transform the dict of queries according to the tuple keys:
    if queries = {
            ('a','b'):'q',
            ('a','c'):'q'
    }
    the results woulf be: {a: {b: 'q', c: 'q'}

    if queries = {
            ('a','b'):'q',
            ('e','c'):'q'
    }
    the results woulf be: {a: {b: 'q'}, e: {c: 'q'}}
    """
    intermediate: t.Dict[str, t.Dict[t.Tuple[str, ...], t.Any]] = defaultdict(
        dict
    )
    final: st.NestedQueryDict = {}
    for query_path, query in queries.items():
        if len(query_path) == 0:
            final[""] = query
        elif len(query_path) == 1:
            final[query_path[0]] = query
        else:
            intermediate[query_path[0]][query_path[1:]] = query

    for name, subdict in intermediate.items():
        final[name] = nest_queries(subdict)

    return final


def nested_dict_of_types(
    types: t.Dict[t.Tuple[str, ...], st.Type],
) -> st.NestedTypeDict:
    """Similar to nest_queries but values are sarus types instead of strings"""
    intermediate: t.Dict[str, t.Dict[t.Tuple[str, ...], t.Any]] = defaultdict(
        dict
    )
    final: st.NestedTypeDict = {}
    for type_path, type in types.items():
        if len(type_path) == 1:
            final[type_path[0]] = type
        else:
            intermediate[type_path[0]][type_path[1:]] = type

    for name, subdict in intermediate.items():
        final[name] = nested_dict_of_types(subdict)

    return final


def nested_unions_from_nested_dict_of_types(
    nested_types: st.NestedTypeDict,
) -> t.Dict[str, st.Type]:
    """create unions out of nested_types"""
    fields: t.Dict[str, st.Type] = {}
    for path_string, type_or_dict in nested_types.items():
        if isinstance(type_or_dict, dict):
            fields[path_string] = sdt.Union(
                nested_unions_from_nested_dict_of_types(type_or_dict)
            )
        else:
            fields[path_string] = t.cast(st.Type, type_or_dict)
    return fields


def translate_query_to_postgres(
    query: str, input_dialect: st.SQLDialect, dataset: pyqrl.Dataset
) -> str:
    """It translates the query to postgres if the input the input dialect is not postgres by using
    sarus_sql.
    TODO: use pyqrlew translator when possible.
    """

    if input_dialect != st.SQLDialect.POSTGRES:
        rel = pyqrl.Relation.from_query(
            query, dataset, SARUS_TO_QRLEW_DIALECT_MAPPING[input_dialect]
        )
        query = rel.to_query(
            SARUS_TO_QRLEW_DIALECT_MAPPING[st.SQLDialect.POSTGRES]
        )
    return query


def translate_queries_to_postgres(
    queries: t.Dict[t.Tuple[str, ...], str],
    input_dialect: st.SQLDialect,
    dataset: pyqrl.Dataset,
) -> t.Dict[t.Tuple[str, ...], str]:
    return {
        path: translate_query_to_postgres(query, input_dialect, dataset)
        for (path, query) in queries.items()
    }



import logging
import typing as t
import warnings
import os
from pathlib import Path
from sarus_data_spec.manager.async_utils import async_iter

try:
    import sarus_llm.recipes.serving as llm_serving
except (ModuleNotFoundError, ValueError, RuntimeError):
    warnings.warn("Sarus LLM not available")

import pyarrow as pa
import sys
from sarus_data_spec.constants import PUBLIC
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.ops.processor.standard.standard_op import (  # noqa: E501
    StandardDatasetImplementation,
    StandardDatasetStaticChecker,
)
from sarus_data_spec.scalar import Scalar
from sarus_data_spec.schema import schema
from sarus_data_spec.type import Struct, Text
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st
import pyarrow.csv as pa_csv

logger = logging.getLogger()


class GenerateFromModelStaticChecker(StandardDatasetStaticChecker):
    async def schema(self) -> st.Schema:
        _, kwargs = self.dataset.parents()
        pretrained_model = kwargs["model"]
        assert isinstance(pretrained_model, Scalar)
        prompts = kwargs["prompts"]
        assert isinstance(prompts, Dataset)
        is_public = pretrained_model.is_public() and prompts.is_public()
        return schema(
            dataset=self.dataset,
            schema_type=Struct(
                {"text": Text()}, properties={PUBLIC: str(is_public)}
            ),
        )


class GenerateFromModel(StandardDatasetImplementation):
    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        _, kwargs = self.dataset.parents()
        pretrained_model = kwargs["model"]
        prompts = kwargs["prompts"]
        transform = self.dataset.transform()

        assert (
            pretrained_model.prototype() == sp.Scalar  # type: ignore # noqa: E501
        ), "pretrained model should be a Scalar"
        pretrained_ds = t.cast(st.Scalar, pretrained_model)

        pretrained = await pretrained_ds.async_value()
        foundation_model_name, checkpoint_path, sample_type = pretrained
        assert prompts.prototype() == sp.Dataset, "prompts should be a Dataset"
        prompts_ds = t.cast(st.Dataset, prompts)

        specs = transform.protobuf().spec.generate_from_model
        dataset = pa.concat_tables(
            [
                pa.Table.from_struct_array(batch.column("sarus_data"))
                async for batch in await prompts_ds.async_to_arrow()
            ]
        )

        # saving_dir
        saving_dir = os.path.join(
            self.dataset.manager().parquet_dir(),
            "llm_sampling",
            self.dataset.uuid(),
        )
        os.makedirs(saving_dir, exist_ok=True)
        # get foundation_model_name
        json_file = os.path.join(saving_dir, "llm_input.jsonl")
        dataset.to_pandas().to_json(json_file, orient="records", lines=True)
        tokenization_dir = os.path.join(
            saving_dir,
            "tokenization",
        )
        sample_saving_dir = os.path.join(saving_dir, "sample.csv")
        cmd, env = prepare_sampling_cmd_env(
            sample_type=sample_type,
            foundation_model_name=foundation_model_name,
            temperature=specs.temperature,
            max_tokens=specs.max_new_tokens,
            restoring_path=checkpoint_path,
            data_dir=json_file,
            tokenization_dir=tokenization_dir,
            sample_saving_dir=sample_saving_dir,
        )
        self.dataset.manager().launch_job(command=cmd, env=env)
        tab = pa_csv.read_csv(sample_saving_dir)
        return async_iter(tab.to_batches(max_chunksize=batch_size))


def prepare_sampling_cmd_env(
    sample_type: str,
    foundation_model_name: str,
    data_dir: str,
    tokenization_dir: str,
    restoring_path: str,
    temperature: float,
    max_tokens: int,
    sample_saving_dir: str,
) -> t.Tuple[t.List[str], t.Dict[str, str]]:
    """
    Prepares and returns the command list and an environment
    with the correct DeepSpeed environment variables.
    """

    cmd = [sys.executable, llm_serving.__file__]
    config_file = str(
        os.path.join(Path(__file__).parent, "sampling_config.yaml")
    )
    cmd.extend(["--config", config_file])
    cmd.append(f"sample_type={sample_type}")
    cmd.append(f"foundation_model_name={foundation_model_name}")
    # dataset info
    cmd.append(f"dataset.data_dir={data_dir}")
    cmd.append(f"dataset.tokenization_dir={tokenization_dir}")
    cmd.append("sampler.triton_kernel=False")
    cmd.append(f"sampler.temperature={temperature}")
    cmd.append(f"sampler.max_length={max_tokens}")
    checkpoint_path = restoring_path if restoring_path != "" else None
    cmd.append(f"checkpoint_path={checkpoint_path}")
    cmd.append(f"saving_dir={sample_saving_dir}")

    current_env = os.environ.copy()
    return cmd, current_env



import base64
from datetime import date, datetime, time, timedelta, timezone
from typing import Any
import json
import sys

from dateutil.parser import parse
import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.ipc as ipc


class SarusJSONEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, np.ndarray):
            return {"_type": "numpy.ndarray", "data": obj.tolist()}
        elif isinstance(obj, pd.DataFrame):
            return {
                "_type": "pandas.DataFrame",
                "data": obj.to_json(date_format="iso"),
            }
        elif isinstance(obj, pd.Series):
            return {
                "_type": "pandas.Series",
                "data": obj.to_json(date_format="iso"),
            }
        elif isinstance(obj, pd.Timestamp):
            return {"_type": "pandas.Timestamp", "data": obj.isoformat()}
        elif isinstance(obj, datetime):
            return {"_type": "datetime", "data": obj.isoformat()}
        elif isinstance(obj, timedelta):
            return {"_type": "timedelta", "data": obj.total_seconds()}
        elif isinstance(obj, timezone):
            utcoffset_result = obj.utcoffset(None)
            if utcoffset_result is not None:
                return {
                    "_type": "timezone",
                    "data": utcoffset_result.total_seconds(),
                }
            else:
                raise ValueError("Invalid timezone object")
        elif isinstance(obj, time):
            return {"_type": "time", "data": obj.isoformat()}
        elif isinstance(obj, date):
            return {"_type": "date", "data": obj.isoformat()}
        elif isinstance(obj, np.generic):
            if np.issubdtype(obj, np.complexfloating):
                complex_obj = obj.astype(np.complex128)
                return {
                    "_type": "numpy.complex",
                    "data": {
                        '"real"': complex_obj.real,
                        '"imag"': complex_obj.imag,
                    },
                }
            else:
                return {"_type": "numpy.generic", "data": obj.item()}
        elif isinstance(obj, pd.MultiIndex):
            return {
                "_type": "pandas.MultiIndex",
                "data": obj.tolist(),
                "names": obj.names,
                "levels": [level.tolist() for level in obj.levels],
                "codes": list(obj.codes),
            }
        elif isinstance(obj, pd.Index):
            return {
                "_type": "pandas.Index",
                "class": type(obj).__name__,
                "data": obj.tolist(),
                "dtype": str(obj.dtype),
            }
        elif isinstance(obj, pd.Period):
            return {
                "_type": "pandas.Period",
                "data": str(obj),
                "freq": obj.freqstr,
            }
        elif isinstance(obj, pd.Timedelta):
            return {"_type": "pandas.Timedelta", "data": obj.value}
        elif isinstance(obj, pd.Interval):
            return {"_type": "pandas.Interval", "data": (obj.left, obj.right)}
        elif isinstance(obj, pd.Categorical):
            return {
                "_type": "pandas.Categorical",
                "data": obj.tolist(),
                "categories": obj.categories.tolist(),
                "ordered": obj.ordered,
            }
        elif isinstance(obj, type):
            return {
                "_type": "class",
                "data": {'"name"': obj.__name__, '"module"': obj.__module__},
            }
        elif isinstance(obj, pd.api.extensions.ExtensionDtype):
            return {"_type": "dtype", "data": str(obj)}
        elif isinstance(obj, slice):
            return {"_type": "slice", "data": (obj.start, obj.stop, obj.step)}
        elif isinstance(obj, range):
            return {
                "_type": "range",
                "data": {
                    '"start"': obj.start,
                    '"stop"': obj.stop,
                    '"step"': obj.step,
                },
            }
        elif isinstance(obj, pa.Table):
            sink = pa.BufferOutputStream()
            writer = ipc.new_stream(sink, obj.schema)
            writer.write_table(obj)
            writer.close()
            serialized_data = (
                sink.getvalue()
            )  # This is the serialized data in a byte array format
            base64_encoded_data = base64.b64encode(serialized_data).decode(
                "utf-8"
            )
            return {"_type": "pa.Table", "data": base64_encoded_data}
        return super().default(obj)

    def encode_obj(self, obj: Any) -> Any:
        if isinstance(obj, tuple):
            return {
                "_type": "tuple",
                "data": [self.encode_obj(v) for v in obj],
            }
        elif isinstance(obj, list):
            return [self.encode_obj(v) for v in obj]
        elif isinstance(obj, dict):
            return {self.encode(k): self.encode_obj(v) for k, v in obj.items()}
        return obj

    def encode(self, obj: Any) -> str:
        obj_transformed = self.encode_obj(obj)
        return super().encode(obj_transformed)

    @classmethod
    def encode_bytes(cls, obj: Any) -> bytes:
        encoder = cls()
        return (encoder.encode(obj)).encode("utf-8")


class SarusJSONDecoder(json.JSONDecoder):
    def decode(self, s: str, *args: Any, **kwargs: Any) -> Any:
        obj = super().decode(s, *args, **kwargs)
        return self.decode_obj(obj)

    def decode_obj(self, obj: Any) -> Any:
        if isinstance(obj, dict):
            if "_type" in obj:
                data = self.decode_obj(obj["data"])
                if obj["_type"] == "tuple":
                    return tuple(self.decode_obj(v) for v in data)
                elif obj["_type"] == "numpy.ndarray":
                    return np.array(data)
                elif obj["_type"] == "pandas.DataFrame":
                    return pd.read_json(data, convert_dates=True)
                elif obj["_type"] == "pandas.Series":
                    return pd.read_json(data, typ="series", convert_dates=True)
                elif obj["_type"] == "pandas.Timestamp":
                    return pd.Timestamp(data)
                elif obj["_type"] == "datetime":
                    return parse(data)
                elif obj["_type"] == "timedelta":
                    return timedelta(seconds=data)
                elif obj["_type"] == "timezone":
                    return timezone(timedelta(seconds=data))
                elif obj["_type"] == "time":
                    return parse(data).time()
                elif obj["_type"] == "date":
                    return parse(data).date()
                elif obj["_type"] == "numpy.generic":
                    return np.array(data).item()
                elif obj["_type"] == "numpy.complex":
                    return np.complex128(complex(data["real"], data["imag"]))
                elif obj["_type"] == "pandas.Index":
                    cls = getattr(pd, obj["class"])
                    return cls(data, dtype=obj["dtype"])
                elif obj["_type"] == "pandas.MultiIndex":
                    return pd.MultiIndex(
                        levels=[pd.Index(level) for level in obj["levels"]],
                        codes=self.decode_obj(obj["codes"]),
                        names=obj["names"],
                    )
                elif obj["_type"] == "pandas.Period":
                    return pd.Period(data, freq=obj["freq"])
                elif obj["_type"] == "pandas.Timedelta":
                    return pd.to_timedelta(data)
                elif obj["_type"] == "pandas.Interval":
                    return pd.Interval(*data)
                elif obj["_type"] == "pandas.Categorical":
                    return pd.Categorical(
                        data,
                        categories=obj["categories"],
                        ordered=obj["ordered"],
                    )
                elif obj["_type"] == "class":
                    if data["module"] in ["builtins", "numpy", "pandas"]:
                        cls = getattr(
                            sys.modules[data["module"]], data["name"]
                        )
                        if isinstance(cls, type):
                            return cls
                        else:
                            raise ValueError("Decoded object is not a type")
                    else:
                        raise ValueError("Invalid module name")
                elif obj["_type"] == "dtype":
                    return pd.api.types.pandas_dtype(data)
                elif obj["_type"] == "slice":
                    return slice(*data)
                elif obj["_type"] == "range":
                    return range(data["start"], data["stop"], data["step"])
                elif obj["_type"] == "pa.Table":
                    binary_data = base64.b64decode(obj["data"])
                    buffer = pa.py_buffer(binary_data)
                    reader = ipc.open_stream(buffer)
                    deserialized_table = reader.read_all()
                    return deserialized_table
            return {self.decode(k): self.decode_obj(v) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [self.decode_obj(v) for v in obj]
        return obj

    @classmethod
    def decode_bytes(cls, b: bytes, *args: Any, **kwargs: Any) -> Any:
        decoder = cls()
        return decoder.decode(b.decode("utf-8"), *args, **kwargs)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/attribute.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n(sarus_data_spec/protobuf/attribute.proto\x12\x0fsarus_data_spec\"\xaa\x01\n\tAttribute\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0e\n\x06object\x18\x02 \x01(\t\x12>\n\nproperties\x18\x03 \x03(\x0b\x32*.sarus_data_spec.Attribute.PropertiesEntry\x12\x0c\n\x04name\x18\x04 \x01(\t\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sarus_data_spec.protobuf.attribute_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ATTRIBUTE_PROPERTIESENTRY._options = None
  _ATTRIBUTE_PROPERTIESENTRY._serialized_options = b'8\001'
  _ATTRIBUTE._serialized_start=62
  _ATTRIBUTE._serialized_end=232
  _ATTRIBUTE_PROPERTIESENTRY._serialized_start=183
  _ATTRIBUTE_PROPERTIESENTRY._serialized_end=232
# @@protoc_insertion_point(module_scope)



import typing
from functools import wraps
from sarus_data_spec.attribute import attach_properties
import sarus_data_spec.typing as st


def check_existence_in_attribute_or_cache_after(
    attribute_name: str,
) -> typing.Callable[
    [typing.Callable[[typing.Any, st.DataSpec], bool]],
    typing.Callable[[typing.Any, st.DataSpec], bool],
]:
    """This decorator is to be used  for methods of
    the dataspec validator that take as input the dataspec
    and return a boolean.
    It first checks if the attribute already exits, and returns it,
    otherwise it executes the method and then caches the result."""

    def cache_and_check(
        func: typing.Callable[[typing.Any, st.DataSpec], bool],
    ) -> typing.Callable[[typing.Any, st.DataSpec], bool]:
        @wraps(func)
        def wrapper_fn(
            validator: typing.Any,
            dataspec: st.DataSpec,
        ) -> bool:
            assert isinstance(dataspec, st.DataSpec)
            existing_attribute = dataspec.attribute(attribute_name)
            if existing_attribute is not None:
                return existing_attribute[attribute_name] == str(True)
            outcome = func(validator, dataspec)
            attach_properties(
                dataspec,
                name=attribute_name,
                properties={attribute_name: str(outcome)},
            )
            return outcome

        return wrapper_fn

    return cache_and_check
